import numpy as np
from numpy import inf
import pandas as pd


def proc_means(data: pd.DataFrame,
                   var: list,
                   precision: int = 2,
                   language: str = 'en') -> pd.DataFrame:
    """
    获取目标数据集的描述统计信息
    Args:
        data:DataFrame,需要进行统计分析的数据集
        var:list,需要统计的变量
        precision:数据精度，小数点位数，默认为2
        language:str,数据结果标题列显示语言，默认为 'en',可手动修改为'cn'

    Returns:
        描述统计结果数据集
    """

    result = []
    for v in var:
        # 数据类型
        col_dtype = str(data[v].dtype)

        # 如果列是数值型，并且不是全为空，则计算统计量
        if col_dtype in ('int64', 'float64', 'float32') and len(data[data[v].notnull()]):
            v_n = data[v].count()
            v_mean = data[v].mean()
            v_std = data[v].std()
            v_min = data[v].min()
            v_max = data[v].max()
            result.append([v, v_n, v_mean, v_std, v_min, v_max])

    result = pd.DataFrame(result, columns=['Variable', 'N', 'Mean', 'Std Dev', 'Minimum', 'Maximum'])

    # 设置显示格式
    result['N'] = result['N'].apply(lambda x: round(x, precision))
    result['Mean'] = result['Mean'].apply(lambda x: round(x, precision))
    result['Std Dev'] = result['Std Dev'].apply(lambda x: round(x, precision))
    result['Minimum'] = result['Minimum'].apply(lambda x: round(x, precision))
    result['Maximum'] = result['Maximum'].apply(lambda x: round(x, precision))

    return result



# -*- coding:utf-8 -*-

# 版本号
__version__ = '0.1.10.1'

import sys
import os
from datetime import datetime
import numpy as np
from numpy import inf
import pandas as pd
import matplotlib.pyplot as plt
from tqdm.notebook import tqdm
import warnings

plt.rcParams['font.sans-serif'] = 'SimHei'  # 设置中文字体
plt.rcParams['axes.unicode_minus'] = False  # 用来正常显示负号

# 导入westat功能模块
# 日志模块
from westat.logger import logger

# 自带数据集
from .dataset import credit_card

# 数据准备 sample
from .sample import read_csv, read_excel, get_data_partition

# 数据探索 explorer
from .explore import (check_data_target,
                      get_data_describe,
                      get_data_distribution,
                      proc_means,
                      )

# 数据处理 modify
from .modify import set_update_bins

# 模型开发 model
from .model import (get_feature_by_ivcorr,
                    get_col_type,
                    get_data_bins,
                    get_data_discrete,
                    get_data_iv,
                    get_data_woe,
                    get_woe_transform,
                    get_model_iv,
                    get_predict_score,
                    get_scorecard,
                    get_tree_bins,
                    get_woe_iv,
                    view_woe_iv,
                    stepwise_forward,
                    stepwise_lr,
                    tree_to_img,
                    tree_to_pdf,
                    )

# 模型评估 access
from .assess import (get_score_distribution,
                     view_score_distribution,
                     get_auc,
                     get_auc_by_card,
                     get_ks,
                     get_ks_by_card,
                     get_vif,
                     get_psi,
                     view_psi,
                     get_data_psi)

# 绘图
from westat.plot_woe import plot_woe
from westat.plot_iv import plot_iv
from westat.plot_corr import plot_corr
from westat.plot_lift import plot_lift
from westat.plot_roc_ks import plot_roc_ks

# 日期处理函数
from westat.date_diff import date_diff

# 设置函数别名
get_data_desc = get_data_describe
get_data_dist = get_data_distribution
get_data_part = get_data_partition
get_score_dist = get_score_distribution
view_score_dist = view_score_distribution

__all__ = (
    'os',
    'sys',
    'datetime',
    'np',
    'pd',
    'inf',
    'plt',
    'tqdm',
    'warnings',
    'utils',
    # 自带数据集
    'credit_card',

    # 日期处理函数
    'date_diff',

    # 数据获取
    'read_csv',
    'read_excel',

    # 数据探索 explorer
    'get_data_partition',
    'get_data_distribution',
    'get_data_describe',
    'proc_means',

    # 数据检查
    'check_data_target',

    # 数据处理 modify
    'get_data_discrete',
    'get_data_iv',
    'get_data_woe',
    'get_woe_transform',
    'get_tree_bins',
    'get_woe_iv',
    'view_woe_iv',
    # dataframe_to_table',
    'get_col_type',
    'get_data_bins',
    'set_update_bins',
    'get_model_iv',

    # 特征筛选
    'get_feature_by_ivcorr',
    'stepwise_lr',
    'stepwise_forward',

    # 模型开发 model
    'get_scorecard',
    'get_predict_score',

    # 模型评估 access
    'get_score_distribution',
    'view_score_distribution',
    'get_auc',
    'get_auc_by_card',
    'get_ks',
    'get_ks_by_card',
    'get_vif',
    'get_psi',
    'view_psi',
    'get_data_psi',

    # 绘图
    'plot_woe',
    'plot_iv',
    'plot_corr',
    'plot_lift',
    'plot_roc_ks',

    # 决策树文件转换
    'tree_to_img',
    'tree_to_pdf',
    # 其他
    'get_data_desc',
    'get_data_dist',
    'get_data_part',
    'get_score_dist',
    'view_score_dist',
)


class Table(pd.DataFrame):
    def __init__(self):
        super().__init__()

    def __str__(self):
        return 'table for westat'



import numpy as np
import pandas as pd
from tqdm.notebook import tqdm
from .get_woe_iv import get_woe_iv


def get_data_woe(data: pd.DataFrame,
                 target='y',
                 method='tree',
                 bins=[],
                 qcut=0,
                 missing: list = [np.nan, None, 'nan'],
                 precision=2):
    """
    批量获取变量WoE值
    Args:
        data: DataFrame,目标数据集
        target: str,目标变量名称，默认为'y'
        method: str,分箱方法，默认为决策树分箱
        bins: list,手动指定的分箱列表
        qcut: int,等额分箱的分组数
        missing: list,缺失值列表
        precision: 数据精度，小数点位数，默认为2

    Returns:
        返回包含特征名称Name，WoE值两列的数据集
    """
    col_iv = []
    for col in tqdm([i for i in data.columns if i != target]):
        if data[col].dtypes in ('int64', 'float64', 'float32'):
            new_method = method
        else:
            new_method = 'discrete'
        col_woe_iv = get_woe_iv(data=data,
                                col=col,
                                target=target,
                                method=new_method,
                                bins=bins,
                                qcut=qcut,
                                missing=missing,
                                show_missing=True,
                                precision=precision)
        col_iv.append([col, float(col_woe_iv['WoE'].sum())])
    result = pd.DataFrame(col_iv, columns=["Name", "WoE"])
    result.sort_values(by='WoE', ascending=False, inplace=True)
    result['WoE'] = result['WoE'].apply(lambda x: round(x, precision))
    return result



# 模型评估 access
from .get_auc import get_auc
from .get_auc_by_card import get_auc_by_card
from .get_data_psi import get_data_psi
from .get_ks import get_ks
from .get_ks_by_card import get_ks_by_card
from .get_psi import get_psi, view_psi
from .get_score_distribution import get_score_distribution, view_score_distribution
from .get_vif import get_vif



import pandas as pd
import os


# 设置自带数据集
def credit_card() -> pd.DataFrame():
    """
    获取自带数据集 uci_credit_card
    Returns:
        返回dataframe格式化的 uci_credit_card 数据集
    """
    from westat.utils import current_path
    file_path = os.path.join(current_path, 'data/UCI_Credit_Card.csv')
    data = pd.read_csv(file_path)
    return data



import numpy as np
import pandas as pd
from tqdm.notebook import tqdm
from .get_woe_iv import get_woe_iv


def get_woe_transform(data_discrete: pd.DataFrame(),
                      fit_transform=pd.DataFrame(),
                      target='y',
                      method='discrete',
                      missing: list = [np.nan, None, 'nan'],
                      precision=2):
    """
    根据离散化数据集，进行WoE转换
    Args:
        data_discrete: DataFrame,目标数据集
        fit_transform:DataFrame,已经离散化的训练集,用于根据训练好的transform，转化test/OOT数据
        target: str,目标变量名称，默认为'y'
        method: str,分箱方法，默认为'discrete'，表示使用已经离散化的数据计算WoE
        missing: list,缺失值列表
        precision: 数据精度，小数点位数，默认为2

    Returns:
        返回经过WoE转换后的数据集
    """
    data_woe = pd.DataFrame()

    if fit_transform.empty:
        for col in tqdm([i for i in data_discrete.columns if i != target]):
            col_woe = get_woe_iv(data_discrete, col=col, target=target, method=method, missing=missing,
                                 precision=precision)
            s = data_discrete[col].replace(list(col_woe['Bin']), list(col_woe['WoE']))
            data_woe = pd.concat([data_woe, s], axis=1)
    else:
        for col in tqdm([i for i in data_discrete.columns if i != target]):
            col_woe = get_woe_iv(fit_transform, col=col, target=target, method=method, missing=missing,
                                 precision=precision)
            s = data_discrete[col].replace(list(col_woe['Bin']), list(col_woe['WoE']))
            data_woe = pd.concat([data_woe, s], axis=1)
    data_woe[target] = data_discrete[target]
    return data_woe



import socket
import getpass
import os

# 设备信息
user_name = getpass.getuser() # 获取当前用户名
host_name = socket.gethostname() # 获取当前主机名
host_ip = socket.gethostbyname(socket.gethostname()) # 获取当前IP

# 路径
current_path = os.path.dirname(os.path.abspath(__file__)) # 获取当前路径



import pandas as pd


def get_vif(data,
            col,
            target='y',
            return_data=False,
            precision=2):
    """
    根据指定的特征清单，计算数据集中每个特征的方差膨胀因子 VIF
    Args:
        data:pd.DataFrame,目标数据集
        col:str，需要计算VIF的特征列表
        target:str,目标变量名称，默认为'y'
        return_data:是否返回结果数据
        precision:int,数据精度，小数点位数，默认为2

    Returns:
        计算 VIF 的值，或根据要求返回结果数据
    """
    from statsmodels.stats.outliers_influence import variance_inflation_factor

    X = data[[i for i in col if i != target]]

    # 计算VIF
    vif = pd.DataFrame()
    vif["variables"] = X.columns
    vif["VIF"] = [variance_inflation_factor(X.values, i) for i in range(X.shape[1])]
    vif_max = round(vif['VIF'].max(), precision)
    vif["VIF"] = vif["VIF"].apply(lambda x: round(x, precision))
    if return_data:
        return vif_max, vif
    else:
        return vif_max



import numpy as np
import pandas as pd
from .get_data_iv import get_data_iv


def get_criterion(lr, crit, x, y):
    if crit == 'aic':
        aic = lr.aic
    return aic


def stepwise_lr(data: pd.DataFrame(),
                target: str = 'y',
                sort_features: list = [],
                crit: str = 'aic',
                verbose: bool = False):
    """
    批量获取变量WoE值
    Args:
        data: 特征变量数据，DataFrame或二维数组
        target:str,目标变量名称，默认为'y'
        sort_features:根据woe排倒序后的特征列表
        crit:str,筛选特征变量的准则，默认为aic
        verbose: 是否输出详细的调试信息，默认为True

    Returns:
        返回特征列表
    """
    import statsmodels.api as sm

    if len(sort_features) == 0:
        data_iv = get_data_iv(data, target=target)
        sort_features = data_iv['Name'].tolist()

    X = data[[col for col in data.columns if col != target]]
    y = data[target]

    # 逐步筛选入模特征
    sort_features_woe = sort_features

    selected_f = [sort_features_woe[0]]
    n = 0
    for f in sort_features_woe[1:]:

        x0 = X[selected_f].copy()
        lr0 = sm.Logit(y, sm.add_constant(x0)).fit(disp=0)

        try_vars = selected_f + [f]
        x = X[try_vars].copy()
        lr = sm.Logit(y, sm.add_constant(x)).fit(disp=0)

        if verbose:
            n = n + 1
            pvals = lr.pvalues.apply(lambda x: round(x, 4))

            print('\nstep {0},current:{1},selected:{3},max_pvalue:{2},pvalue:{1},pvalues:{5}'.format(n, f, selected_f,
                                                                                                     pvals[-1],
                                                                                                     max(pvals),
                                                                                                     pvals.tolist()))
        # 判断新变量是否显著, 如果不显著则跳过
        if lr.pvalues[-1] >= 0.05:
            if verbose:
                print('新变量不显著')
            continue
        # 如果所有变量都显著，则判断目标函数值是否下降，如果下降则加入变量到候选变量
        elif max(lr.pvalues) < 0.05:
            if verbose:
                print('所有变量都显著')
            if get_criterion(lr, crit, x, y) < get_criterion(lr0, crit, x0, y):
                selected_f.append(f)
        # 如果新变量显著，且有老变量不显著，则判断比较两者的损失函数值下降水平，保留更优的
        else:
            if verbose:
                print('新变量显著，且有老变量不显著')
            del try_vars[np.argmax(lr.pvalues)]
            x1 = X[try_vars].copy()
            lr1 = sm.Logit(y, sm.add_constant(x1)).fit(disp=0)
            # 如果去掉不显著变量后目标函数下降，则入选新特征
            aic1 = round(get_criterion(lr1, crit, x1, y), 4)
            aic2 = round(get_criterion(lr0, crit, x0, y), 4)
            if aic1 < aic2:
                if verbose:
                    print('去掉不显著变量后目标函数下降，去掉后的aic:{},去掉前的aic:{}'.format(aic1, aic2))
                    print(np.argmax(lr.pvalues))
                    print(selected_f[np.argmax(lr.pvalues) - 1])
                del selected_f[np.argmax(lr.pvalues) - 1]
                selected_f.append(f)

    return selected_f



import numpy as np
import pandas as pd
import statsmodels.api as sm


def stepwise_forward(X, y, initial_list=[], threshold_in=0.01, threshold_out=0.05, verbose=True):
    """
    使用逐步向前法实现逻辑回归特征选择。

    参数：
    X: 特征变量数据，DataFrame或二维数组
    y: 目标变量数据，Series或一维数组
    initial_list: 初始特征列表，默认为空
    threshold_in: 添加特征的p值阈值，默认为0.01
    threshold_out: 移除特征的p值阈值，默认为0.05
    verbose: 是否输出详细的调试信息，默认为True

    返回值：
    selected_features: 选中的特征列表
    """

    included = list(initial_list)
    while True:
        changed = False
        excluded = list(set(X.columns) - set(included))

        # 添加特征
        new_pval = pd.Series(index=excluded, dtype=np.float64)
        for new_column in excluded:
            model = sm.Logit(y, sm.add_constant(X[included + [new_column]])).fit(disp=0)
            new_pval[new_column] = model.pvalues[new_column]
        best_pval = new_pval.min()
        if best_pval < threshold_in:
            best_feature = new_pval.idxmin()
            included.append(best_feature)
            changed = True
            if verbose:
                print(f'Add  {best_feature} with p-value {best_pval}')

        # 移除特征
        model = sm.Logit(y, sm.add_constant(X[included])).fit(disp=0)
        p_values = model.pvalues.iloc[1:]
        worst_pval = p_values.max()
        if worst_pval > threshold_out:
            changed = True
            worst_feature = p_values.idxmax()
            included.remove(worst_feature)
            if verbose:
                print(f'Drop {worst_feature} with p-value {worst_pval}')

        if not changed:
            break

    return included



import numpy as np
from numpy import inf
import pandas as pd

from .get_psi import get_psi


def get_data_psi(data_actual: pd.DataFrame,
                 data_expected: str,
                 qcut=10,
                 missing: list = [np.nan, None, 'nan'],
                 target='y',
                 precision=6) -> pd.DataFrame:
    """
    计算数据集PSI
    Args:
        data_actual: DataFrame,实际数据集
        data_expected: DataFrame,预期数据集
        col:str,需要计算PSI的列名
        bins:list，计算PSI的分箱
        qcut: int,等额分箱的分组数
        missing: list,缺失值列表
        target: str,目标变量名称，默认为'y'
        precision:数据精度，小数点位数，默认为2
        language: str,数据结果标题列显示语言，默认为 'en',可手动修改为'cn'

    Returns:
        结果数据集保存数据集对应的PSI
    """
    col_psi = []
    for col in [i for i in data_actual.columns if i != target]:
        df_psi = get_psi(data_actual,
                         data_expected,
                         col,
                         qcut=qcut,
                         target=target,
                         missing=missing,
                         precision=precision)
        col_psi.append([col, float(df_psi['Total PSI'].loc[0])])

    result = pd.DataFrame(col_psi, columns=["Name", "PSI"])
    result.sort_values(by='PSI', ascending=False, inplace=True)
    return result



# 模型开发
from .feature_selection import get_feature_by_ivcorr
from .get_col_type import get_col_type
from .get_data_bins import get_data_bins
from .get_data_discrete import get_data_discrete
from .get_data_iv import get_data_iv
from .get_data_woe import get_data_woe
from .get_woe_transform import get_woe_transform
from .get_model_iv import get_model_iv
from .get_predict_score import get_predict_score
from .get_scorecard import get_scorecard
from .get_tree_bins import get_tree_bins
from .get_woe_iv import get_woe_iv, view_woe_iv
from .stepwise_forward import stepwise_forward
from .stepwise_lr import stepwise_lr

# 决策树
from .tree import tree_to_img, tree_to_pdf



import numpy as np
import pandas as pd


def get_ks(data: pd.DataFrame,
           col: str = 'Proba',
           target: str = 'y',
           qcut: int = 10,
           method: str = '',
           return_data: bool = False,
           precision: int = 2):
    """
    根据评分卡内容，对目标数据集的评分结果计算 KS
    Args:
        data:pd.DataFrame,目标数据集
        col:str，正例的概率估计字段名
        target:tr,目标变量名称，默认为'y'
        qcut:int,分组数，默认切分为10组
        method:str,计算ks的方法，默认为空，可选sklearn
        return_data:是否返回结果数据
        precision:int,数据精度，小数点位数，默认为2

    Returns:
        计算ks的值，或根据要求返回结果数据
    """

    if method == 'sklearn':
        from sklearn.metrics import roc_curve
        fpr, tpr, thresholds = roc_curve(data[target], data[col], drop_intermediate=False)
        pre = sorted(data[col], reverse=True)
        num = [i * int(len(pre) / qcut) for i in range(qcut)]
        num = num + [(len(pre) - 1)]
        ks_thresholds = [max(thresholds[thresholds <= pre[i]]) for i in num]
        result = pd.DataFrame([fpr, tpr, thresholds, tpr - fpr]).T
        result.columns = ['fpr', 'tpr', 'thresholds', 'ks']
        result = pd.merge(result, pd.DataFrame(ks_thresholds, columns=['thresholds']), on='thresholds', how='inner')
        result.reset_index(drop=True, inplace=True)
        result['No.'] = result.index + 1
        result = result[['No.', 'fpr', 'tpr', 'thresholds', 'ks']]

        # 数据显示格式处理
        result['fpr'] = result['fpr'].apply(lambda x: round(x, precision))
        result['tpr'] = result['tpr'].apply(lambda x: round(x, precision))
        result['ks'] = result['ks'].apply(lambda x: round(abs(x), precision))
        result = result.head(qcut)
        ks = round(result['ks'].max(), precision)
    else:
        df = data.copy()
        df[col] = pd.qcut(x=df[col], q=qcut, duplicates='drop')
        total = df.groupby([col])[target].count()
        bad = df.groupby([col])[target].sum()
        result = pd.DataFrame({'#Total': total, '#Bad': bad})
        result['#Good'] = result['#Total'] - result['#Bad']
        result['Proba'] = result.index
        result.index = range(len(result))
        result = result.sort_values(by='Proba', ascending=True)
        result['%Bad'] = result['#Bad'] / result['#Bad'].sum()
        result['%Good'] = result['#Good'] / result['#Good'].sum()
        result['%Total'] = result['#Total'] / result['#Total'].sum()
        result['%BadRate'] = result['#Bad'] / result['#Total']

        result['%CumBad'] = result['#Bad'].cumsum() / result['#Bad'].sum()
        result['%CumGood'] = result['#Good'].cumsum() / result['#Good'].sum()
        result['KS'] = result['%CumGood'] - result['%CumBad']
        result.reset_index(drop=True, inplace=True)
        result['No.'] = result.index + 1
        result = result[
            ['No.', 'Proba', '#Total', '#Bad', '#Good', '%Total', '%Bad', '%Good', '%BadRate', '%CumBad', '%CumGood',
             'KS']]

        result['KS'] = result['KS'].apply(lambda x: round(abs(x), precision))
        result['%Total'] = result['%Total'].apply(lambda x: format(x, '.' + str(precision) + '%'))
        result['%Bad'] = result['%Bad'].apply(lambda x: format(x, '.' + str(precision) + '%'))
        result['%Good'] = result['%Good'].apply(lambda x: format(x, '.' + str(precision) + '%'))
        result['%BadRate'] = result['%BadRate'].apply(lambda x: format(x, '.' + str(precision) + '%'))
        result['%CumBad'] = result['%CumBad'].apply(lambda x: format(x, '.' + str(precision) + '%'))
        result['%CumGood'] = result['%CumGood'].apply(lambda x: format(x, '.' + str(precision) + '%'))
        ks = round(result['KS'].max(), precision)

    if return_data:
        return ks, result

    else:
        return ks
